from .arch import Arch, Endness, register_arch
from .archerror import ArchError
from .tls import TLSArchInfo


class ArchAVR8(Arch):
    bits = 32
    vex_arch = None
    name = "AVR8"
    qemu_name = "avr"
    linux_name = "avr"  # ???
    triplet = "avr-linux-gnu"  # ???
    max_inst_bytes = 4
    instruction_alignment = 2
    elf_tls = TLSArchInfo(1, 8, [], [0], [], 0, 0)  # ???

    def __init__(self, endness=Endness.LE):
        if endness != Endness.LE:
            raise ArchError("Arch AVR8 must be little endian")
        super().__init__(endness)

        # IO Registers are mapped into the register file starting at 0x20
        # Any instruction that references an IO register by numer should just add this.
        self.ioreg_offset = 0x20
        # Instructions and data are in different memory in AVR
        # Translate data address into the address space by adding this
        self.data_offset = 0x10000000

        self.registers = {}
        self.registers.update({"r%d" % i: (i, 1) for i in range(0, 32)})
        self.registers.update({"R%d_R%d" % (i + 1, i): (i, 2) for i in range(0, 32, 2)})

        self.registers["W"] = (24, 2)
        self.registers["X"] = (26, 2)
        self.registers["Y"] = (28, 2)
        self.registers["Z"] = (30, 2)

        self.registers["WL"] = (24, 1)
        self.registers["WH"] = (25, 1)
        self.registers["XL"] = (26, 1)
        self.registers["XH"] = (27, 1)
        self.registers["YL"] = (28, 1)
        self.registers["YH"] = (29, 1)
        self.registers["ZL"] = (30, 1)
        self.registers["ZH"] = (31, 1)

        self.registers["EEDR"] = (0x40, 1)
        self.registers["EEARL"] = (0x41, 1)
        self.registers["EEARH"] = (0x42, 1)
        self.registers["GTCCR"] = (0x43, 1)
        self.registers["TCCR0A"] = (0x44, 1)
        self.registers["TCCR0B"] = (0x45, 1)
        self.registers["TCNT0"] = (0x46, 1)
        self.registers["OCR0A"] = (0x47, 1)
        self.registers["OCR0B"] = (0x48, 1)
        self.registers["IO_0x29"] = (0x49, 1)
        self.registers["GPIOR1"] = (0x4A, 1)
        self.registers["GPIOR2"] = (0x4B, 1)
        self.registers["SPCR"] = (0x4C, 1)
        self.registers["SPSR"] = (0x4D, 1)
        self.registers["SPDR"] = (0x4E, 1)
        self.registers["IO_0x2f"] = (0x4F, 1)
        self.registers["ACSR"] = (0x50, 1)
        self.registers["IO_0x31"] = (0x51, 1)
        self.registers["IO_0x32"] = (0x52, 1)
        self.registers["SMCR"] = (0x53, 1)
        self.registers["MCUSR"] = (0x54, 1)
        self.registers["MCUCR"] = (0x55, 1)
        self.registers["IO_0x36"] = (0x56, 1)
        self.registers["SPMCSR"] = (0x57, 1)
        self.registers["RAMPD"] = (0x58, 1)
        self.registers["RAMPX"] = (0x59, 1)
        self.registers["RAMPY"] = (0x5A, 1)
        self.registers["RAMPZ"] = (0x5B, 1)
        self.registers["EIND"] = (0x5C, 1)

        self.registers["SP"] = (0x5D, 2)
        self.registers["SPL"] = (0x5D, 1)
        self.registers["SPH"] = (0x5E, 1)

        self.registers["SREG"] = (0x5F, 1)

        self.registers["WDTCSR"] = (0x60, 1)
        self.registers["CLKPR"] = (0x61, 1)
        self.registers["PRR"] = (0x64, 1)
        self.registers["OSCCAL"] = (0x66, 1)
        self.registers["PCICR"] = (0x68, 1)
        self.registers["EICRA"] = (0x69, 1)
        self.registers["PCMSK0"] = (0x6B, 1)
        self.registers["PCMSK2"] = (0x6D, 1)
        self.registers["PCMSK1"] = (0x6C, 1)
        self.registers["TIMSK0"] = (0x6E, 1)
        self.registers["TIMSK1"] = (0x6F, 1)
        self.registers["TIMSK2"] = (0x70, 1)
        self.registers["ADCL"] = (0x78, 1)
        self.registers["ADCH"] = (0x79, 1)
        self.registers["ADCSRA"] = (0x7A, 1)
        self.registers["ADCSRB"] = (0x7B, 1)
        self.registers["ADMUX"] = (0x7C, 1)
        self.registers["DIDR0"] = (0x7E, 1)
        self.registers["DIDR1"] = (0x7F, 1)
        self.registers["TCCR1A"] = (0x80, 1)
        self.registers["TCCR1B"] = (0x81, 1)
        self.registers["TCCR1C"] = (0x82, 1)
        self.registers["TCNT1H"] = (0x85, 1)
        self.registers["TCNT1L"] = (0x84, 1)
        self.registers["ICR1H"] = (0x87, 1)
        self.registers["ICR1L"] = (0x86, 1)
        self.registers["OCR1AH"] = (0x89, 1)
        self.registers["OCR1AL"] = (0x88, 1)
        self.registers["OCR1BH"] = (0x8B, 1)
        self.registers["OCR1BL"] = (0x8A, 1)
        self.registers["TCCR2A"] = (0xB0, 1)
        self.registers["TCCR2B"] = (0xB1, 1)
        self.registers["TCNT2"] = (0xB2, 1)
        self.registers["OCR2A"] = (0xB3, 1)
        self.registers["OCR2B"] = (0xB4, 1)
        self.registers["ASSR"] = (0xB6, 1)
        self.registers["TWBR"] = (0xB8, 1)
        self.registers["TWSR"] = (0xB9, 1)
        self.registers["TWAR"] = (0xBA, 1)
        self.registers["TWDR"] = (0xBB, 1)
        self.registers["TWCR"] = (0xBC, 1)
        self.registers["TWAMR"] = (0xBD, 1)
        self.registers["UCSR0A"] = (0xC0, 1)
        self.registers["UCSR0B"] = (0xC1, 1)
        self.registers["UCSR0C"] = (0xC2, 1)
        self.registers["UBRR0H"] = (0xC5, 1)
        self.registers["UBRR0L"] = (0xC4, 1)
        self.registers["UDR0"] = (0xC6, 1)

        self.registers["pc"] = (0x80000000, 2)
        self.registers["ip"] = (0x80000000, 2)
        self.registers["sp"] = self.registers["SP"]

        self.register_names = {}
        self.register_names.update({i: "r%d" % i for i in range(0, 32)})
        self.register_names[self.registers["pc"][0]] = "pc"
        self.register_names[self.registers["sp"][0]] = "sp"

        self.ip_offset = self.registers["pc"][0]
        self.sp_offset = self.registers["sp"][0]


register_arch([r".*avr"], 32, Endness.LE, ArchAVR8)



from archinfo.types import RegisterOffset

from .arch import Arch, Endness, Register, register_arch
from .tls import TLSArchInfo

try:
    import capstone as _capstone
except ImportError:
    _capstone = None

try:
    import keystone as _keystone
except ImportError:
    _keystone = None

try:
    import unicorn as _unicorn
    from unicorn import riscv_const
except ImportError:
    _unicorn = None


def is_riscv_arch(a):
    return a.name.startswith("RISCV")


class ArchRISCV64(Arch):
    def __init__(self, endness=Endness.LE):
        super().__init__(endness)
        if endness == Endness.BE:
            self.function_prologs = set()
            self.function_epilogs = set()

    bits = 64
    vex_arch = "VexArchRISCV64"
    name = "RISCV64"
    qemu_name = "riscv64"
    linux_name = "riscv64"
    triplet = "riscv64-linux-gnu"
    max_inst_bytes = 4
    ret_offset = RegisterOffset(4)
    vex_conditional_helpers = True
    syscall_num_offset = 132
    call_pushes_ret = False
    memory_endness = Endness.LE
    register_endness = Endness.LE
    instruction_endness = Endness.LE
    sizeof = {"short": 16, "int": 32, "long": 64, "long long": 64}
    if _capstone and hasattr(_capstone, "CS_ARCH_RISCV"):
        cs_arch = _capstone.CS_ARCH_RISCV
        cs_mode = _capstone.CS_MODE_RISCV64
    # if _keystone:
    #     ks_arch = _keystone.KS_ARCH_RISCV
    #     ks_mode = _keystone.KS_MODE_LITTLE_ENDIAN
    uc_arch = _unicorn.UC_ARCH_RISCV if _unicorn else None
    uc_mode = _unicorn.UC_MODE_RISCV64 if _unicorn else None
    uc_const = riscv_const if _unicorn else None
    uc_prefix = "UC_RISCV_" if _unicorn else None

    ret_instruction = b"\x00\x00\x80\x67"  # ret
    nop_instruction = b"\x00\x00\x00\x13"  # nop
    function_prologs = set()
    function_epilogs = set()
    instruction_alignment = 4
    register_list = [
        Register(name="x0", size=8, alias_names=("zero",)),
        Register(
            name="x1",
            size=8,
            alias_names=(
                "ra",
                "lr",
            ),
            general_purpose=True,
        ),
        Register(name="x2", size=8, alias_names=("sp",), general_purpose=True),
        Register(name="x3", size=8, alias_names=("gp",), general_purpose=True),
        Register(name="x4", size=8, alias_names=("tp",), general_purpose=True),
        Register(name="x5", size=8, alias_names=("t0",), general_purpose=True),
        Register(name="x6", size=8, alias_names=("t1",), general_purpose=True),
        Register(name="x7", size=8, alias_names=("t2",), general_purpose=True),
        Register(name="x8", size=8, alias_names=("s0", "fp", "bp"), general_purpose=True),
        Register(name="x9", size=8, alias_names=("s1",), general_purpose=True),
        Register(name="x10", size=8, alias_names=("a0",), general_purpose=True, argument=True),
        Register(name="x11", size=8, alias_names=("a1",), general_purpose=True, argument=True),
        Register(name="x12", size=8, alias_names=("a2",), general_purpose=True, argument=True),
        Register(name="x13", size=8, alias_names=("a3",), general_purpose=True, argument=True),
        Register(name="x14", size=8, alias_names=("a4",), general_purpose=True, argument=True),
        Register(name="x15", size=8, alias_names=("a5",), general_purpose=True, argument=True),
        Register(name="x16", size=8, alias_names=("a6",), general_purpose=True, argument=True),
        Register(name="x17", size=8, alias_names=("a7",), general_purpose=True, argument=True),
        Register(name="x18", size=8, alias_names=("s2",), general_purpose=True),
        Register(name="x19", size=8, alias_names=("s3",), general_purpose=True),
        Register(name="x20", size=8, alias_names=("s4",), general_purpose=True),
        Register(name="x21", size=8, alias_names=("s5",), general_purpose=True),
        Register(name="x22", size=8, alias_names=("s6",), general_purpose=True),
        Register(name="x23", size=8, alias_names=("s7",), general_purpose=True),
        Register(name="x24", size=8, alias_names=("s8",), general_purpose=True),
        Register(name="x25", size=8, alias_names=("s9",), general_purpose=True),
        Register(name="x26", size=8, alias_names=("s10",), general_purpose=True),
        Register(name="x27", size=8, alias_names=("s11",), general_purpose=True),
        Register(name="x28", size=8, alias_names=("t3",), general_purpose=True),
        Register(name="x29", size=8, alias_names=("t4",), general_purpose=True),
        Register(name="x30", size=8, alias_names=("t5",), general_purpose=True),
        Register(name="x31", size=8, alias_names=("t6",), general_purpose=True),
        Register(name="pc", size=8, alias_names=("ip",)),
        Register(
            name="f0",
            size=8,
            alias_names=("ft0",),
            floating_point=True,
        ),
        Register(name="f1", size=8, alias_names=("ft1",), floating_point=True),
        Register(name="f2", size=8, alias_names=("ft2",), floating_point=True),
        Register(name="f3", size=8, alias_names=("ft3",), floating_point=True),
        Register(name="f4", size=8, alias_names=("ft4",), floating_point=True),
        Register(name="f5", size=8, alias_names=("ft5",), floating_point=True),
        Register(name="f6", size=8, alias_names=("ft6",), floating_point=True),
        Register(name="f7", size=8, alias_names=("ft7",), floating_point=True),
        Register(name="f8", size=8, alias_names=("fs0",), floating_point=True),
        Register(name="f9", size=8, alias_names=("fs1",), floating_point=True),
        Register(name="f10", size=8, alias_names=("fa0",), floating_point=True),
        Register(name="f11", size=8, alias_names=("fa1",), floating_point=True),
        Register(name="f12", size=8, alias_names=("fa2",), floating_point=True),
        Register(name="f13", size=8, alias_names=("fa3",), floating_point=True),
        Register(name="f14", size=8, alias_names=("fa4",), floating_point=True),
        Register(name="f15", size=8, alias_names=("fa5",), floating_point=True),
        Register(name="f16", size=8, alias_names=("fa6",), floating_point=True),
        Register(name="f17", size=8, alias_names=("fa7",), floating_point=True),
        Register(name="f18", size=8, alias_names=("fs2",), floating_point=True),
        Register(name="f19", size=8, alias_names=("fs3",), floating_point=True),
        Register(name="f20", size=8, alias_names=("fs4",), floating_point=True),
        Register(name="f21", size=8, alias_names=("fs5",), floating_point=True),
        Register(name="f22", size=8, alias_names=("fs6",), floating_point=True),
        Register(name="f23", size=8, alias_names=("fs7",), floating_point=True),
        Register(name="f24", size=8, alias_names=("fs8",), floating_point=True),
        Register(name="f25", size=8, alias_names=("fs9",), floating_point=True),
        Register(name="f26", size=8, alias_names=("fs10",), floating_point=True),
        Register(name="f27", size=8, alias_names=("fs11",), floating_point=True),
        Register(name="f28", size=8, alias_names=("ft8",), floating_point=True),
        Register(name="f29", size=8, alias_names=("ft9",), floating_point=True),
        Register(name="f30", size=8, alias_names=("ft10",), floating_point=True),
        Register(name="f31", size=8, alias_names=("ft11",), floating_point=True),
        Register(name="ip_at_syscall", size=8, artificial=True),
    ]

    got_section_name = ".got"
    ld_linux_name = "ld-linux-riscv64-lp64d.so.1"
    elf_tls = TLSArchInfo(1, 32, [], [0], [], 0, 0)


register_arch([r".*riscv.*"], 64, Endness.ANY, ArchRISCV64)



try:
    import capstone as _capstone
except ImportError:
    _capstone = None

try:
    import keystone as _keystone
except ImportError:
    _keystone = None

try:
    import pyvex as _pyvex
except ImportError:
    _pyvex = None

from archinfo.types import RegisterOffset

from .arch import Arch, Endness, Register, register_arch
from .archerror import ArchError
from .tls import TLSArchInfo


class ArchS390X(Arch):
    def __init__(self, endness=Endness.BE):
        super().__init__(endness)
        if endness != Endness.BE:
            raise ArchError("Arch s390x must be big endian")
        self.argument_register_positions = (
            {
                self.registers["r2"][0]: 0,
                self.registers["r3"][0]: 1,
                self.registers["r4"][0]: 2,
                self.registers["r5"][0]: 3,
                self.registers["r6"][0]: 4,
                # fp registers
                self.registers["f0"][0]: 0,
                self.registers["f2"][0]: 1,
                self.registers["f4"][0]: 2,
                self.registers["f6"][0]: 3,
            }
            if _pyvex is not None
            else None
        )

    bits = 64
    vex_arch = "VexArchS390X"  # enum VexArch
    name = "S390X"
    default_endness = Endness.BE
    qemu_name = "s390x"  # target/s390x
    triplet = "s390x-linux-gnu"
    linux_name = "s390"  # arch/s390
    max_inst_bytes = 6
    ret_offset = RegisterOffset(584)  # offsetof(VexGuestS390XState, guest_r2)
    syscall_num_offset = 576  # offsetof(VexGuestS390XState, guest_r1)
    call_pushes_ret = False
    stack_change = -8
    initial_sp = 0x40000000000
    sizeof = {"short": 16, "int": 32, "long": 64, "long long": 64}
    if _capstone:
        cs_arch = _capstone.CS_ARCH_SYSZ
        cs_mode = _capstone.CS_MODE_BIG_ENDIAN
    if _keystone:
        ks_arch = _keystone.KS_ARCH_SYSTEMZ
        ks_mode = _keystone.KS_MODE_BIG_ENDIAN
    ret_instruction = b"\x07\xf4"  # br %r14
    nop_instruction = b"\x07\x07"  # nopr %r7
    instruction_alignment = 2
    register_list = [
        Register(name="ia", size=8, alias_names=("ip", "pc")),
        Register(name="r0", size=8, general_purpose=True),
        Register(name="r1", size=8, general_purpose=True, subregisters=[("r1_32", 4, 4)]),
        Register(name="r2", size=8, general_purpose=True, argument=True, subregisters=[("r2_32", 4, 4)]),
        Register(
            name="r3",
            size=8,
            general_purpose=True,
            argument=True,
            linux_entry_value="argc",
            subregisters=[("r3_32", 4, 4)],
        ),
        Register(
            name="r4",
            size=8,
            general_purpose=True,
            argument=True,
            linux_entry_value="argv",
            subregisters=[("r4_32", 4, 4)],
        ),
        Register(
            name="r5",
            size=8,
            general_purpose=True,
            argument=True,
            linux_entry_value="envp",
            subregisters=[("r5_32", 4, 4)],
        ),
        Register(
            name="r6", size=8, general_purpose=True, argument=True, persistent=True, subregisters=[("r6_32", 4, 4)]
        ),
        Register(name="r7", size=8, general_purpose=True, persistent=True, subregisters=[("r7_32", 4, 4)]),
        Register(name="r8", size=8, general_purpose=True, persistent=True, subregisters=[("r8_32", 4, 4)]),
        Register(name="r9", size=8, general_purpose=True, persistent=True, subregisters=[("r9_32", 4, 4)]),
        Register(name="r10", size=8, general_purpose=True, persistent=True, subregisters=[("r10_32", 4, 4)]),
        Register(
            name="r11",
            size=8,
            alias_names=("bp",),
            general_purpose=True,
            persistent=True,
            subregisters=[("r11_32", 4, 4)],
        ),
        Register(name="r12", size=8, general_purpose=True, persistent=True, subregisters=[("r12_32", 4, 4)]),
        Register(name="r13", size=8, general_purpose=True, persistent=True, subregisters=[("r13_32", 4, 4)]),
        # Strictly speaking, there is no fixed link register on s390x.
        # However, %r14 is almost always used for that, so mark it as such.
        # Situations when that's not the case (e.g. brasl %r0,X)
        # can still be handled explicitly.
        Register(name="r14", size=8, general_purpose=True, alias_names=("lr",)),
        Register(
            name="r15",
            size=8,
            alias_names=("sp",),
            general_purpose=True,
            persistent=True,
            default_value=(initial_sp, True, "global"),
        ),
        Register(name="v0", size=16, subregisters=[("f0", 0, 8)], floating_point=True),
        Register(name="v1", size=16, subregisters=[("f1", 0, 8)], floating_point=True),
        Register(name="v2", size=16, subregisters=[("f2", 0, 8)], floating_point=True),
        Register(name="v3", size=16, subregisters=[("f3", 0, 8)], floating_point=True),
        Register(name="v4", size=16, subregisters=[("f4", 0, 8)], floating_point=True),
        Register(name="v5", size=16, subregisters=[("f5", 0, 8)], floating_point=True),
        Register(name="v6", size=16, subregisters=[("f6", 0, 8)], floating_point=True),
        Register(name="v7", size=16, subregisters=[("f7", 0, 8)], floating_point=True),
        Register(name="v8", size=16, subregisters=[("f8", 0, 8)], floating_point=True),
        Register(name="v9", size=16, subregisters=[("f9", 0, 8)], floating_point=True),
        Register(name="v10", size=16, subregisters=[("f10", 0, 8)], floating_point=True),
        Register(name="v11", size=16, subregisters=[("f11", 0, 8)], floating_point=True),
        Register(name="v12", size=16, subregisters=[("f12", 0, 8)], floating_point=True),
        Register(name="v13", size=16, subregisters=[("f13", 0, 8)], floating_point=True),
        Register(name="v14", size=16, subregisters=[("f14", 0, 8)], floating_point=True),
        Register(name="v15", size=16, subregisters=[("f15", 0, 8)], floating_point=True),
        Register(name="v16", size=16, vector=True),
        Register(name="v17", size=16, vector=True),
        Register(name="v18", size=16, vector=True),
        Register(name="v19", size=16, vector=True),
        Register(name="v20", size=16, vector=True),
        Register(name="v21", size=16, vector=True),
        Register(name="v22", size=16, vector=True),
        Register(name="v23", size=16, vector=True),
        Register(name="v24", size=16, vector=True),
        Register(name="v25", size=16, vector=True),
        Register(name="v26", size=16, vector=True),
        Register(name="v27", size=16, vector=True),
        Register(name="v28", size=16, vector=True),
        Register(name="v29", size=16, vector=True),
        Register(name="v30", size=16, vector=True),
        Register(name="v31", size=16, vector=True),
        Register(name="a0", size=4),
        Register(name="a1", size=4),
        Register(name="a2", size=4),
        Register(name="a3", size=4),
        Register(name="a4", size=4),
        Register(name="a5", size=4),
        Register(name="a6", size=4),
        Register(name="a7", size=4),
        Register(name="a8", size=4),
        Register(name="a9", size=4),
        Register(name="a10", size=4),
        Register(name="a11", size=4),
        Register(name="a12", size=4),
        Register(name="a13", size=4),
        Register(name="a14", size=4),
        Register(name="a15", size=4),
        Register(name="nraddr", size=8),
        Register(name="cmstart", size=8),
        Register(name="cmlen", size=8),
        Register(name="ip_at_syscall", size=8, artificial=True),
        Register(name="emnote", size=4, artificial=True),
    ]

    function_prologs = {
        rb"\xeb.[\xf0-\xff]..\x24",  # stmg %r1,%r3,d2(%r15)
    }
    function_epilogs = {
        rb"\x07\xf4",  # br %r14
    }

    got_section_name = ".got"
    ld_linux_name = "ld64.so.1"
    elf_tls = TLSArchInfo(
        variant=2,  # 3.4.7 @ https://www.uclibc.org/docs/tls.pdf
        tcbhead_size=64,  # sizeof(tcbhead_t)
        head_offsets=[0],  # offsetof(tcbhead_t, tcb)
        dtv_offsets=[8],  # offsetof(tcbhead_t, dtv)
        pthread_offsets=[16],  # offsetof(tcbhead_t, self)
        tp_offset=0,
        dtv_entry_offset=0,
    )

    dwarf_registers = [
        "v0",
        "v1",
        "v2",
        "v3",
        "v4",
        "v5",
        "v6",
        "v7",
        "v8",
        "v9",
        "v10",
        "v11",
        "v12",
        "v13",
        "v14",
        "v15",
        "v16",
        "v17",
        "v18",
        "v19",
        "v20",
        "v21",
        "v22",
        "v23",
        "v24",
        "v25",
        "v26",
        "v27",
        "v28",
        "v29",
        "v30",
        "v31",
        "cr0",
        "cr1",
        "cr2",
        "cr3",
        "cr4",
        "cr5",
        "cr6",
        "cr7",
        "cr8",
        "cr9",
        "cr10",
        "cr11",
        "cr12",
        "cr13",
        "cr14",
        "cr15",
        "ar0",
        "ar1",
        "ar2",
        "ar3",
        "ar4",
        "ar5",
        "ar6",
        "ar7",
        "ar8",
        "ar9",
        "ar10",
        "ar11",
        "ar12",
        "ar13",
        "ar14",
        "ar15",
        "psw_mask",
        "psw_address",
    ]


register_arch(["s390"], 64, Endness.BE, ArchS390X)



from archinfo.types import RegisterOffset

from .arch import Arch, Endness, Register, register_arch
from .tls import TLSArchInfo

try:
    import capstone as _capstone
except ImportError:
    _capstone = None

try:
    import keystone as _keystone
except ImportError:
    _keystone = None

try:
    import unicorn as _unicorn
except ImportError:
    _unicorn = None


class ArchAArch64(Arch):
    def __init__(self, endness=Endness.LE):
        super().__init__(endness)
        if endness == Endness.BE:
            self.ida_processor = "armb"
            self.function_prologs = set()
            self.function_epilogs = set()

    bits = 64
    vex_arch = "VexArchARM64"
    name = "AARCH64"
    qemu_name = "aarch64"
    ida_processor = "arm"
    linux_name = "aarch64"
    triplet = "aarch64-linux-gnueabihf"
    max_inst_bytes = 4
    ret_offset = RegisterOffset(16)
    vex_conditional_helpers = True
    syscall_num_offset = 80
    call_pushes_ret = False
    stack_change = -8
    memory_endness = Endness.LE
    register_endness = Endness.LE
    instruction_endness = Endness.LE
    sizeof = {"short": 16, "int": 32, "long": 64, "long long": 64}
    if _capstone:
        cs_arch = _capstone.CS_ARCH_ARM64
        cs_mode = _capstone.CS_MODE_LITTLE_ENDIAN
    if _keystone:
        ks_arch = _keystone.KS_ARCH_ARM64
        ks_mode = _keystone.KS_MODE_LITTLE_ENDIAN
    uc_arch = _unicorn.UC_ARCH_ARM64 if _unicorn else None
    uc_mode = _unicorn.UC_MODE_LITTLE_ENDIAN if _unicorn else None
    uc_const = _unicorn.arm64_const if _unicorn else None
    uc_prefix = "UC_ARM64_" if _unicorn else None
    initial_sp = 0x7FFFFFFFFFF0000

    ret_instruction = b"\xc0\x03\x5f\xd6"  # ret
    nop_instruction = b"\x1f\x20\x03\xd5"  # nop
    instruction_alignment = 4
    register_list = [
        Register(
            name="x0",
            size=8,
            subregisters=[("w0", 0, 4)],
            alias_names=("r0",),
            general_purpose=True,
            argument=True,
            linux_entry_value="ld_destructor",
        ),
        Register(
            name="x1", size=8, subregisters=[("w1", 0, 4)], alias_names=("r1",), general_purpose=True, argument=True
        ),
        Register(
            name="x2", size=8, subregisters=[("w2", 0, 4)], alias_names=("r2",), general_purpose=True, argument=True
        ),
        Register(
            name="x3", size=8, subregisters=[("w3", 0, 4)], alias_names=("r3",), general_purpose=True, argument=True
        ),
        Register(
            name="x4", size=8, subregisters=[("w4", 0, 4)], alias_names=("r4",), general_purpose=True, argument=True
        ),
        Register(
            name="x5", size=8, subregisters=[("w5", 0, 4)], alias_names=("r5",), general_purpose=True, argument=True
        ),
        Register(
            name="x6", size=8, subregisters=[("w6", 0, 4)], alias_names=("r6",), general_purpose=True, argument=True
        ),
        Register(name="x7", size=8, subregisters=[("w7", 0, 4)], alias_names=("r7",), general_purpose=True),
        Register(name="x8", size=8, subregisters=[("w8", 0, 4)], alias_names=("r8",), general_purpose=True),
        Register(name="x9", size=8, subregisters=[("w9", 0, 4)], alias_names=("r9",), general_purpose=True),
        Register(name="x10", size=8, subregisters=[("w10", 0, 4)], alias_names=("r10",), general_purpose=True),
        Register(name="x11", size=8, subregisters=[("w11", 0, 4)], alias_names=("r11",), general_purpose=True),
        Register(name="x12", size=8, subregisters=[("w12", 0, 4)], alias_names=("r12",), general_purpose=True),
        Register(name="x13", size=8, subregisters=[("w13", 0, 4)], alias_names=("r13",), general_purpose=True),
        Register(name="x14", size=8, subregisters=[("w14", 0, 4)], alias_names=("r14",), general_purpose=True),
        Register(name="x15", size=8, subregisters=[("w15", 0, 4)], alias_names=("r15",), general_purpose=True),
        Register(name="x16", size=8, subregisters=[("w16", 0, 4)], alias_names=("r16", "ip0"), general_purpose=True),
        Register(name="x17", size=8, subregisters=[("w17", 0, 4)], alias_names=("r17", "ip1"), general_purpose=True),
        Register(name="x18", size=8, subregisters=[("w18", 0, 4)], alias_names=("r18",), general_purpose=True),
        Register(name="x19", size=8, subregisters=[("w19", 0, 4)], alias_names=("r19",), general_purpose=True),
        Register(name="x20", size=8, subregisters=[("w20", 0, 4)], alias_names=("r20",), general_purpose=True),
        Register(name="x21", size=8, subregisters=[("w21", 0, 4)], alias_names=("r21",), general_purpose=True),
        Register(name="x22", size=8, subregisters=[("w22", 0, 4)], alias_names=("r22",), general_purpose=True),
        Register(name="x23", size=8, subregisters=[("w23", 0, 4)], alias_names=("r23",), general_purpose=True),
        Register(name="x24", size=8, subregisters=[("w24", 0, 4)], alias_names=("r24",), general_purpose=True),
        Register(name="x25", size=8, subregisters=[("w25", 0, 4)], alias_names=("r25",), general_purpose=True),
        Register(name="x26", size=8, subregisters=[("w26", 0, 4)], alias_names=("r26",), general_purpose=True),
        Register(name="x27", size=8, subregisters=[("w27", 0, 4)], alias_names=("r27",), general_purpose=True),
        Register(name="x28", size=8, subregisters=[("w28", 0, 4)], alias_names=("r28",), general_purpose=True),
        Register(
            name="x29", size=8, subregisters=[("w29", 0, 4)], alias_names=("r29", "fp", "bp"), general_purpose=True
        ),
        Register(name="x30", size=8, subregisters=[("w30", 0, 4)], alias_names=("r30", "lr"), general_purpose=True),
        Register(
            name="xsp",
            size=8,
            subregisters=[("wsp", 0, 4)],
            alias_names=("sp",),
            general_purpose=True,
            default_value=(initial_sp, True, "global"),
        ),
        Register(name="pc", size=8, alias_names=("ip",)),
        Register(name="cc_op", size=8, artificial=True),
        Register(name="cc_dep1", size=8, artificial=True),
        Register(name="cc_dep2", size=8, artificial=True),
        Register(name="cc_ndep", size=8, artificial=True),
        Register(name="tpidr_el0", size=8),
        Register(
            name="q0",
            size=16,
            subregisters=[("d0", 0, 8), ("s0", 0, 4), ("h0", 0, 2), ("b0", 0, 1)],
            alias_names=("v0",),
            floating_point=True,
            vector=True,
        ),
        Register(
            name="q1",
            size=16,
            subregisters=[("d1", 0, 8), ("s1", 0, 4), ("h1", 0, 2), ("b1", 0, 1)],
            alias_names=("v1",),
            floating_point=True,
            vector=True,
        ),
        Register(
            name="q2",
            size=16,
            subregisters=[("d2", 0, 8), ("s2", 0, 4), ("h2", 0, 2), ("b2", 0, 1)],
            alias_names=("v2",),
            floating_point=True,
            vector=True,
        ),
        Register(
            name="q3",
            size=16,
            subregisters=[("d3", 0, 8), ("s3", 0, 4), ("h3", 0, 2), ("b3", 0, 1)],
            alias_names=("v3",),
            floating_point=True,
            vector=True,
        ),
        Register(
            name="q4",
            size=16,
            subregisters=[("d4", 0, 8), ("s4", 0, 4), ("h4", 0, 2), ("b4", 0, 1)],
            alias_names=("v4",),
            floating_point=True,
            vector=True,
        ),
        Register(
            name="q5",
            size=16,
            subregisters=[("d5", 0, 8), ("s5", 0, 4), ("h5", 0, 2), ("b5", 0, 1)],
            alias_names=("v5",),
            floating_point=True,
            vector=True,
        ),
        Register(
            name="q6",
            size=16,
            subregisters=[("d6", 0, 8), ("s6", 0, 4), ("h6", 0, 2), ("b6", 0, 1)],
            alias_names=("v6",),
            floating_point=True,
            vector=True,
        ),
        Register(
            name="q7",
            size=16,
            subregisters=[("d7", 0, 8), ("s7", 0, 4), ("h7", 0, 2), ("b7", 0, 1)],
            alias_names=("v7",),
            floating_point=True,
            vector=True,
        ),
        Register(
            name="q8",
            size=16,
            subregisters=[("d8", 0, 8), ("s8", 0, 4), ("h8", 0, 2), ("b8", 0, 1)],
            alias_names=("v8",),
            floating_point=True,
            vector=True,
        ),
        Register(
            name="q9",
            size=16,
            subregisters=[("d9", 0, 8), ("s9", 0, 4), ("h9", 0, 2), ("b9", 0, 1)],
            alias_names=("v9",),
            floating_point=True,
            vector=True,
        ),
        Register(
            name="q10",
            size=16,
            subregisters=[("d10", 0, 8), ("s10", 0, 4), ("h10", 0, 2), ("b10", 0, 1)],
            alias_names=("v10",),
            floating_point=True,
            vector=True,
        ),
        Register(
            name="q11",
            size=16,
            subregisters=[("d11", 0, 8), ("s11", 0, 4), ("h11", 0, 2), ("b11", 0, 1)],
            alias_names=("v11",),
            floating_point=True,
            vector=True,
        ),
        Register(
            name="q12",
            size=16,
            subregisters=[("d12", 0, 8), ("s12", 0, 4), ("h12", 0, 2), ("b12", 0, 1)],
            alias_names=("v12",),
            floating_point=True,
            vector=True,
        ),
        Register(
            name="q13",
            size=16,
            subregisters=[("d13", 0, 8), ("s13", 0, 4), ("h13", 0, 2), ("b13", 0, 1)],
            alias_names=("v13",),
            floating_point=True,
            vector=True,
        ),
        Register(
            name="q14",
            size=16,
            subregisters=[("d14", 0, 8), ("s14", 0, 4), ("h14", 0, 2), ("b14", 0, 1)],
            alias_names=("v14",),
            floating_point=True,
            vector=True,
        ),
        Register(
            name="q15",
            size=16,
            subregisters=[("d15", 0, 8), ("s15", 0, 4), ("h15", 0, 2), ("b15", 0, 1)],
            alias_names=("v15",),
            floating_point=True,
            vector=True,
        ),
        Register(
            name="q16",
            size=16,
            subregisters=[("d16", 0, 8), ("s16", 0, 4), ("h16", 0, 2), ("b16", 0, 1)],
            alias_names=("v16",),
            floating_point=True,
            vector=True,
        ),
        Register(
            name="q17",
            size=16,
            subregisters=[("d17", 0, 8), ("s17", 0, 4), ("h17", 0, 2), ("b17", 0, 1)],
            alias_names=("v17",),
            floating_point=True,
            vector=True,
        ),
        Register(
            name="q18",
            size=16,
            subregisters=[("d18", 0, 8), ("s18", 0, 4), ("h18", 0, 2), ("b18", 0, 1)],
            alias_names=("v18",),
            floating_point=True,
            vector=True,
        ),
        Register(
            name="q19",
            size=16,
            subregisters=[("d19", 0, 8), ("s19", 0, 4), ("h19", 0, 2), ("b19", 0, 1)],
            alias_names=("v19",),
            floating_point=True,
            vector=True,
        ),
        Register(
            name="q20",
            size=16,
            subregisters=[("d20", 0, 8), ("s20", 0, 4), ("h20", 0, 2), ("b20", 0, 1)],
            alias_names=("v20",),
            floating_point=True,
            vector=True,
        ),
        Register(
            name="q21",
            size=16,
            subregisters=[("d21", 0, 8), ("s21", 0, 4), ("h21", 0, 2), ("b21", 0, 1)],
            alias_names=("v21",),
            floating_point=True,
            vector=True,
        ),
        Register(
            name="q22",
            size=16,
            subregisters=[("d22", 0, 8), ("s22", 0, 4), ("h22", 0, 2), ("b22", 0, 1)],
            alias_names=("v22",),
            floating_point=True,
            vector=True,
        ),
        Register(
            name="q23",
            size=16,
            subregisters=[("d23", 0, 8), ("s23", 0, 4), ("h23", 0, 2), ("b23", 0, 1)],
            alias_names=("v23",),
            floating_point=True,
            vector=True,
        ),
        Register(
            name="q24",
            size=16,
            subregisters=[("d24", 0, 8), ("s24", 0, 4), ("h24", 0, 2), ("b24", 0, 1)],
            alias_names=("v24",),
            floating_point=True,
            vector=True,
        ),
        Register(
            name="q25",
            size=16,
            subregisters=[("d25", 0, 8), ("s25", 0, 4), ("h25", 0, 2), ("b25", 0, 1)],
            alias_names=("v25",),
            floating_point=True,
            vector=True,
        ),
        Register(
            name="q26",
            size=16,
            subregisters=[("d26", 0, 8), ("s26", 0, 4), ("h26", 0, 2), ("b26", 0, 1)],
            alias_names=("v26",),
            floating_point=True,
            vector=True,
        ),
        Register(
            name="q27",
            size=16,
            subregisters=[("d27", 0, 8), ("s27", 0, 4), ("h27", 0, 2), ("b27", 0, 1)],
            alias_names=("v27",),
            floating_point=True,
            vector=True,
        ),
        Register(
            name="q28",
            size=16,
            subregisters=[("d28", 0, 8), ("s28", 0, 4), ("h28", 0, 2), ("b28", 0, 1)],
            alias_names=("v28",),
            floating_point=True,
            vector=True,
        ),
        Register(
            name="q29",
            size=16,
            subregisters=[("d29", 0, 8), ("s29", 0, 4), ("h29", 0, 2), ("b29", 0, 1)],
            alias_names=("v29",),
            floating_point=True,
            vector=True,
        ),
        Register(
            name="q30",
            size=16,
            subregisters=[("d30", 0, 8), ("s30", 0, 4), ("h30", 0, 2), ("b30", 0, 1)],
            alias_names=("v30",),
            floating_point=True,
            vector=True,
        ),
        Register(
            name="q31",
            size=16,
            subregisters=[("d31", 0, 8), ("s31", 0, 4), ("h31", 0, 2), ("b31", 0, 1)],
            alias_names=("v31",),
            floating_point=True,
            vector=True,
        ),
        Register(name="qcflag", size=16, floating_point=True),
        Register(name="emnote", size=4, artificial=True),
        Register(name="cmstart", size=8),
        Register(name="cmlen", size=8),
        Register(name="nraddr", size=8),
        Register(name="ip_at_syscall", size=8, artificial=True),
        Register(name="fpcr", size=4, floating_point=True, default_value=(initial_sp, True, "global")),
    ]

    got_section_name = ".got"
    ld_linux_name = "ld-linux-aarch64.so.1"
    elf_tls = TLSArchInfo(1, 32, [], [0], [], 0, 0)
    dwarf_registers = [
        "x0",
        "x1",
        "x2",
        "x3",
        "x4",
        "x5",
        "x6",
        "x7",
        "x8",
        "x9",
        "x10",
        "x11",
        "x12",
        "x13",
        "x14",
        "x15",
        "x16",
        "x17",
        "x18",
        "x19",
        "x20",
        "x21",
        "x22",
        "x23",
        "x24",
        "x25",
        "x26",
        "x27",
        "x28",
        "x29",
        "x30",
        "sp",
        "<none>",
        "ELR_mode",
        "RA_SIGN_STATE",
        "<none>",
        "<none>",
        "<none>",
        "<none>",
        "<none>",
        "<none>",
        "<none>",
        "<none>",
        "<none>",
        "<none>",
        "<none>",
        "VG",
        "FFR",
        "p0",
        "p1",
        "p2",
        "p3",
        "p4",
        "p5",
        "p6",
        "p7",
        "p8",
        "p9",
        "p10",
        "p11",
        "p12",
        "p13",
        "p14",
        "p15",
        "v0",
        "v1",
        "v2",
        "v3",
        "v4",
        "v5",
        "v6",
        "v7",
        "v8",
        "v9",
        "v10",
        "v11",
        "v12",
        "v13",
        "v14",
        "v15",
        "v16",
        "v17",
        "v18",
        "v19",
        "v20",
        "v21",
        "v22",
        "v23",
        "v24",
        "v25",
        "v26",
        "v27",
        "v28",
        "v29",
        "v30",
        "v31",
        "z0",
        "z1",
        "z2",
        "z3",
        "z4",
        "z5",
        "z6",
        "z7",
        "z8",
        "z9",
        "z10",
        "z11",
        "z12",
        "z13",
        "z14",
        "z15",
        "z16",
        "z17",
        "z18",
        "z19",
        "z20",
        "z21",
        "z22",
        "z23",
        "z24",
        "z25",
        "z26",
        "z27",
        "z28",
        "z29",
        "z30",
        "z31",
    ]


register_arch([r".*arm64.*|.*aarch64*"], 64, Endness.ANY, ArchAArch64)
